use anyhow::bail;
use anyhow::Context;
use anyhow::Result;
use tempfile::NamedTempFile;

use std::env;
use std::ffi::{OsStr, OsString};
use std::fs;
use std::fs::File;
use std::io::{self, Read, Seek, SeekFrom, Write};
use std::path::Path;

// Offset within the bank file to the version number within the BKHD section.
// TODO: Check if this may change.
const OFFSET_VERSION_NUMBER: u64 = 52;
// Retrieved by philipdestroyer by inspecting the game's sound banks in memory at runtime.
const BANK_VERSION: [u8; 4] = [0x28, 0xbc, 0x11, 0x92];

fn fix_file<P>(path: P) -> Result<()>
where
    P: AsRef<Path>,
{
    let tmp_path = {
        let mut f = File::open(&path).with_context(|| "failed to open sound bank file")?;
        let mut tmp = NamedTempFile::new().with_context(|| "failed to open temporary file")?;

        // Copy everything until the version number
        io::copy(
            &mut std::io::Read::by_ref(&mut f).take(OFFSET_VERSION_NUMBER),
            &mut tmp,
        )
        .with_context(|| "failed to copy to tempfile")?;

        // Write the hard coded version number
        tmp.write(&BANK_VERSION)
            .with_context(|| "failed to write version number")?;

        // Copy the rest of the file
        f.seek(SeekFrom::Start(OFFSET_VERSION_NUMBER + 4))
            .with_context(|| "failed to skip to rest of content")?;
        io::copy(&mut f, &mut tmp).with_context(|| "failed to copy to tempfile")?;
        tmp.into_temp_path()
    };

    if let Err(_) = fs::rename(&tmp_path, &path) {
        // Attempt manual copy-and-delete
        let mut tmp = File::open(&tmp_path).with_context(|| "failed to open tempfile for reading")?;
        let mut f = File::open(&path).with_context(|| "failed to open destination file")?;
        if let Err(err) = io::copy(&mut tmp, &mut f) {
            let tmp_path = tmp_path
                .keep()
                .with_context(|| "failed to persist temporary file")?;
            bail!(
                "failed to rename or move temporary file to override the sound bank: {}!\n\
                Please manually move the file from {} to {}.",
                err,
                tmp_path.display(),
                path.as_ref().display()
            );
        }
    }

    Ok(())
}

fn main() -> Result<()> {
    let dir = env::args()
        .nth(1)
        .expect("Please specify a directory containing .wwise_bank files");

    let mut results = Vec::new();

    for entry in fs::read_dir(dir)? {
        let entry = entry?;
        let path = entry.path();

        if path.is_file() && path.extension().and_then(OsStr::to_str) == Some("wwise_bank") {
            let res = fix_file(&path);
            results.push((path, res));
        }
    }

    for (path, res) in results.iter() {
        let name = path
            .file_name()
            .map(OsStr::to_os_string)
            .unwrap_or_else(|| OsString::from(".."));

        match res {
            Ok(_) => println!("Fixed: {}", name.to_string_lossy()),
            Err(err) => eprintln!("Failed: {}. Error: {}", name.to_string_lossy(), err),
        }
    }

    Ok(())
}
